--[[
==================================================================
Name:		  LibFrameX-1.0
Author(s): 	  Azmaedus
Revision: 	  $Rev: 5 $
Created:	  01/23/2015
Description:  Library that creates a frame structure dynamically.
Dependencies: LibStub
License:	  GPLv2
==================================================================
Change Log
--------------------
01/23/2015 - 1.0a 		Creation of addon
01/24/2015 - 1.0a-r2	Changed the name from GJBPlayer-1.0 to LibFrameX-1.0
01/24/2015 - 1.0a-r3	Optimization pass and documentation

==================================================================
Structure of the FrameX.tframes table
------------------------------------------------------------------
- Each table within tframes can have the following elements:
root: 		This frame's elements wont be drawn; acts as an invisible container |boolean|(optional)
ftype: 		Type of frame (Frame, Button, ...) |string|
name: 		Given name of the frame |string|
dim:		Dimensions |dimension table: w,h,s|(width, height, scale)
childdim:	Dimensions of all element's children |dimension table: w,h,s|(width, height, scale)
pos: 		Position |position table: x, y, a, r, ap| (x, y, anchor, relative to, r anchor)
show:		Visibility property |bolean|
parentrgba:	This element uses it's parent childrgba values |boolean|
parentbrgba:This element uses it's parent childbrgba values |boolean|
parentbd:	This element uses it's parent backdrop values |boolean|
parentdim:	This element uses it's parent dimension values |boolean|
parenttext:	This element uses it's parent text values |boolean|
parentname:	Name of the parent element |string|

	---------------------------------------------
	||Text||
text: 		Table of text properties to show a string of characters in the frame |text table: str,color,fs,halign,valign|(optional)
	str: 	Text to be displayed |string|
	color:	Colour of the text |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)
	fs: 	Font style |string| defaults to: "NumberFont_Shadow_Small"
	halign: horizontal alignment |string| defaults to: "CENTER"
	valign: vertical alignment |string| defaults to: "CENTER"
	show:	show/hide text |boolean|
	
childtext:	Table of text properties to show a string of characters in the elements children's frame |text table: str,color,fs,halign,valign|(optional)
	str: 	Text to be displayed |string|
	color:	Colour of the text |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)
	fs: 	Font style |string| defaults to: "NumberFont_Shadow_Small"
	halign: horizontal alignment |string| defaults to: "CENTER"
	valign: vertical alignment |string| defaults to: "CENTER"
	show:	show/hide text |boolean|

	---------------------------------------------
	||Backdrops||
	A backdrop is the background part of a frame; it can be colourized
	Here is the structure:
	bgFile: Background image file; such as Interface\ChatFrame\ChatFrameBackground
	edgeFile: Edge file of the background image; such as Interface/Tooltips/UI-Tooltip-Border
	tile: Is the image tiled or stretched |boolean|
	tileSize: Size in pixels of each tile |integer|
	edgeSize: Size of the edge and corners |integer|
	insets: The bigger the edge, the bigger the inset. Controls overflow, making sure backdrop doesn't overflow outside of the edges.
		{ left, right, top, bottom } |integer|

bd: 		BackDrop of the element |BackDrop table|(optional)
childbd:	BackDrop of all child elements of this element |BarckDrop table|(optional)

	---------------------------------------------
	||Colours||
rgba: 		Colour of the backdrop |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)
rgbashaded:	Swap colour of the backdrop |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)
childrgba:	Colour of the all child backdrops |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)
childrgbashaded:	Swap colour of all the child backdrops |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)
brgba: 		Colour of the element's border |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)
brgbashaded:Swap colour of the element's border |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)
childbrgba:	Colour of the element's child borders |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)
childbrgbashaded:	Swap colour of the element's child borders |RGBA table:0-255,0-255,0-255,1-255,0-1|(optional)

	---------------------------------------------
	||Movement & Scripts||
movable: Makes the frame movable with a specific mouse button and script. |movable table: enabled, drag table|(optional)
	enabled: Enable or disable the drag |boolean| 
	drag: List of buttons to register for the drag |drag table: str-1,...,str-n| ("LeftButton", "RightButton", ...)

clicks: List of possible registered mouse clicks |click table: str-1, ..., str-n| ("LeftButton", "RightButton", ...)
scripts: Table of scripts associated with this element |scripts table: event, func|
	event: Event name |string| ("OnSomething")
	func: actual function to execute when event is triggered |function(self,...) ... end|
==================================================================
]]

local LIBFRAMEX = "LibFrameX-1.0"
local LIBFRAMEX_MINOR = 0 + tonumber(("$Revision: 5 $"):match("(%d+)"))

local FrameX = LibStub:NewLibrary(LIBFRAMEX, LIBFRAMEX_MINOR)
if not FrameX then return end

FrameX.tframes = {} 		-- FrameX table of frames
FrameX.prev = UIParent 		-- Storage for previous frame inside of FrameX.tframes

-- Default settings for FrameX.frame's creation
local defaults = {
	-- Background
	bg = {
		-- Backdrop info
		bd = {
			bgFile = [[Interface\ChatFrame\ChatFrameBackground]],
			edgeFile = [[Interface/Tooltips/UI-Tooltip-Border]],
			tile = true, tileSize = 4, edgeSize = 4,
			insets = { left = 1, right = 1, top = 1, bottom = 1 }
		},
		align = "CENTER", xPos = 0, yPos = 0, Scale = 1, w = 200,	h = 150, r = 0,	g = 0, b = 0, a = 1, br = 1, bg = 1, bb = 1, ba = 1,
	},
}

-- creation of the main lib frame
if not FrameX.frame then
	FrameX.frame = CreateFrame("Frame")
	local f, bg = FrameX.frame, defaults.bg
	f:SetWidth(bg.w)
	f:SetHeight(bg.h)
	f:SetBackdrop(bg.bd)
	f:SetBackdropColor(bg.r, bg.g, bg.b, bg.a)
	f:SetBackdropBorderColor(bg.br, bg.bg, bg.bb, bg.ba)
	f.pos = {}
	f:Show()
end

------------------------------------------------------------
-- Localization enUS
------------------------------------------------------------
local L_FRAMEX_TITLE = "LibFrameX-1.0 Frame"
local L_FRAMEX_ERROR = "Error creating frame!"

local l = GetLocale()
if l == "frFR" then
	L_FRAMEX_TITLE = "LibFrameX-1.0 Frame"
	L_FRAMEX_ERROR = "Erreur lors de la creation de la frame!"
elseif l == "koKR" then
	-- Missing
elseif l == "deDE" then
	-- Missing
elseif l == "zhCN" then
	-- Missing
elseif l == "zhTW" then
	-- Missing
elseif l == "ruRU" then
	-- Missing
elseif l == "esES" or l == "esMX" then
	-- Missing
elseif l == "ptBR" then
	-- Missing
elseif l == "itIT" then
	-- Missing
end

------------------------------------------------------------
-- findFrame: searches for a frame inside of FrameX.tframes and returns the index
------------------------------------------------------------
-- Returns the index of a frame in LFX.tframes by its name
function FrameX:findFrame(name)
	t = self.tframes
	for i = 1, #t do
		if t[i].name == name then
			return i
		end
	end
	return nil
end

------------------------------------------------------------
-- Function to create a frame with values passed from FrameX:makeFrames
------------------------------------------------------------
function FrameX:genFrame(f, parent)
	if not f then return end -- fail
	
	local fo = CreateFrame(f.ftype, f.name, parent, f.template)
	fo.name = f.name
	fo.parentrgba = f.parentrgba
	fo.parentdim = f.parentdim
	fo.parentname = f.parentname
	fo.childrgba = f.childrgba
	fo.childrgbashaded = f.childrgbashaded
	fo.childbrgba = f.childbrgba
	fo.childbd = f.childbd
	fo.childdim = f.childdim
	fo.parenttext = f.parenttext
	fo.rgba = f.rgba
	fo.rgbashaded = f.rgbashaded
	fo.brgba = f.brgba
	fo.bd = f.bd
	fo.dim = f.dim
	fo.root = f.root
	fo.text = f.text
	fo.childtext = f.childtext
	
	local pid = self:findFrame(f.parentname)
	
	if f.parentname and (f.parentrgba or f.parentbrgba or f.parentbd) then
		local rgba = {}
		local brgba = {}
		local bd = {}
		if parent and not f.root then
			if pid then
				pf = self.tframes[pid]
				rgba = pf.childrgba or pf.rgba or f.rgba
				brgba = pf.childbrgba or pf.brgba or f.brgba
				bd = pf.childbd or pf.bd or f.bd
			else
				rgba = f.rgba
				brgba = f.brgba
				bd = f.bd
			end
		else
			rgba = f.rgba
			brgba = f.brgba
			bd = f.bd
		end
		fo:SetBackdrop(bd)
		fo:SetBackdropColor(rgba.r/256, rgba.g/256, rgba.b/256, rgba.a)
		fo:SetBackdropBorderColor(brgba.r/256, brgba.g/256, brgba.b/256, brgba.a)
	elseif f.bd then
		fo:SetBackdrop(f.bd)
		fo:SetBackdropColor(f.rgba.r/256, f.rgba.g/256, f.rgba.b/256, f.rgba.a)
		fo:SetBackdropBorderColor(f.brgba.r/256, f.brgba.g/256, f.brgba.b/256, f.brgba.a)
	end

	if f.parentdim then
		local dim = {}
		if parent then
			if pid then
				dim = self.tframes[pid].childdim
			else
				dim = f.dim
			end
		else
			dim = f.dim
		end
		fo:SetWidth(dim.w)
		fo:SetHeight(dim.h)
		fo:SetScale(dim.s)
	else
		fo:SetWidth(f.dim.w)
		fo:SetHeight(f.dim.h)
		fo:SetScale(f.dim.s)
	end

	if f.bd then
	end
	
	if f.show then
		fo:Show()
	end
	return fo
end

function FrameX:makeFrames(t)
FrameX.tframes = {}
	for i = 1, #t do
		v = t[i]
		-- check if frame exists
		if v then
			self.tframes[i] = self:genFrame(v, self.prev) -- create the frame
			xf = self.tframes[i]
			self.prev = xf -- set previous frame object
			xf:SetPoint(v.pos.a, v.pos.r, v.pos.rp, v.pos.x, v.pos.y) -- align, relative, rel parent, x, y
			
			-- check if the frame has a text component
			if (v.text and v.text.show) or v.parenttext then
				txt = {}
				if v.text then
					txt = v.text
				elseif v.parenttext then
					local id = self:findFrame(v.parentname)
					if id then
						txt = self.tframes[id].childtext
					else
						txt = v.text
					end
				end
				if txt.show then
					xf.text = xf:CreateFontString(nil, "ARTWORK", txt.fs or "NumberFont_Shadow_Small")
					xf.text:SetAllPoints(true)
					xf.text:SetJustifyH(txt.halign or "CENTER")
					xf.text:SetJustifyV(txt.valign or "CENTER")
					xf.text:SetTextColor(txt.color.r/256 or 1, txt.color.g/256 or 1, txt.color.b/256 or 1, txt.color.a or 1)
					xf.text:SetText(txt.str or "")
				end
			end

			--TODO: set texture
			--FrameX.tframes[i]:SetNormalTexture("\\media\\button.tga")
			
			-- if frame is movable
			if v.movable and v.movable.enabled then
				xf:SetMovable(true)
				xf:EnableMouse(true)
				xf:RegisterForDrag(unpack(v.movable.drag))
				--xf.text:SetScale(v.dim.s) ????? why di I put this here
			else
				xf:SetMovable(false)
				xf:EnableMouse(false)
			end

			-- register click events
			if v.clicks then
				FrameX.tframes[i]:SetPoint("CENTER")
				FrameX.tframes[i]:RegisterForClicks(unpack(v.clicks))
			end
			
			-- create scripts
			if v.scripts then -- be careful, if you have scripts, you should have a register clicks
				for y = 1, #v.scripts do
					s = v.scripts[y]
					xf:SetScript(s.event, s.func)
				end
			end
		else
			assert(false, L_FRAMEX_ERROR .. " #" .. i) -- Error
		end
	end
end

function FrameX:setVisible(v)
	f = self.tframes
	for i = 1, #t do
		if v then
			f[i]:Show()
		else
			f[i]:Hide()
		end
	end
end

function FrameX:setFrameVisible(i, s)
	f = self.tframes
	if s then
		f[i]:Show()
	else
		f[i]:Hide()
	end
end

function FrameX:setPosition(i, a, r, rp, x, y)
	f = self.tframes
	f[i]:SetPoint(a, r, rp, x, y)
end

function FrameX:setFrameStr(i, str)
	txt = self.tframes[i].text
	if txt.show then
		txt:SetText(str)
	end
end